# Parse SPLConqueror output logs. Requires Python 3.
import numpy as np
import pandas as pd
import itertools
import os


class Model:

    def __init__(self, model_expression=None, model_terms=None):
        if model_expression is not None:
            self.terms = list()
            terms = model_expression.split(' + ')
            for term in terms:
                influence, *combination = term.split(' * ')
                influence = float(influence.strip())
                self.terms.append(ModelTerm(influence, combination))
        elif model_terms is not None:
            self.terms = model_terms
        else:
            raise ValueError('You have to provide either a mdoel expression or a list of model terms.')
        self.size = sum([t.size for t in self.terms])

    def __repr__(self):
        return ' + '.join([str(t) for t in self.terms])

    def predict(self, configuration):
        prediction = 0 if len(self.terms) else None
        for term in self.terms:
            term_combination = frozenset(term.combination)
            if term_combination.issubset(['root']):

                # In the models produced by SPLConqueror, the root feature (if
                # it exists in the variability model) is always renamed in
                # "root".  So, the influence of the "root" feature from the
                # influence model is always added.
                prediction += term.influence
                continue
            if term_combination.issubset(configuration):
                prediction += term.influence
        return prediction


class ModelTerm:

    def __init__(self, influence, combination):
        self.influence = influence
        self.combination = combination
        self.size = len(self.combination)

    def __repr__(self):
        return str(self.influence) + ' * ' + ' * '.join(sorted(self.combination))


def model(model_expression=None, model_terms=None):
    return Model(model_expression, model_terms)


def parse_df(pathname, log_format='latest'):
    
    # Calculate how many lines schould be skpped in the file until the lines
    # with learning round statistics.
    csv_skipfooter = 0
    with open(pathname) as log_file:
        for num, line in enumerate(log_file, 1):
            if line.startswith('1;'):
                csv_skip = num - 1
            if line.startswith('Analyze finished') and num > csv_skip:
                csv_skipfooter = 1

    # pandas cannot interpret the value used to represet
    # infinity. float_converter is used to convert the infinity value to the
    # maximum float number if it occures in the dataset.
    float_converter = lambda x: np.float64(x)
    if log_format == 'metastudy':
        csv_header = ['Round', 'Model', 'LearningError', 'LearningErrorRel',
                      'ValidationError', 'ValidationErrorRel', 'ElapsedSeconds',
                      'ModelComplexity', 'BestCandidate', 'BestCandidateSize',
                      'BestCandidateScore', 'BestCandidatePenalizedScore',
                      'SomeError']

        csv_converters = {'Model':model, 'BestCandidate':model,
                      'ErrorScore':float_converter,
                      'PenalizedErrorScore':float_converter}
        #csv_skip = 16; # Skip log header till the lines with lerning round stats
    elif log_format == 'scores':
        csv_header = ['Round', 'Model', 'LearningError', 'LearningErrorRel',
                      'ValidationError', 'ValidationErrorRel', 'ElapsedSeconds',
                      'ModelComplexity', 'BestCandidate', 'BestCandidateSize',
                      'ErrorScore', 'PenalizedErrorScore', 'InfluenceScore',
                      'PenalizedInfluenceScore', 'SomeError']

        csv_converters = {'Model':model, 'BestCandidate':model,
                      'ErrorScore':float_converter,
                      'PenalizedErrorScore':float_converter,
                      'InfluenceScore':float_converter,
                      'PenalizedInfluenceScore':float_converter}
        # csv_skip = 17; # Skip log header till the lines with lerning round stats
    else:
        csv_header = ['Round', 'Model', 'LearningError', 'LearningErrorRel',
                      'ValidationError', 'ValidationErrorRel', 'ElapsedSeconds',
                      'ModelComplexity', 'BestCandidate', 'BestCandidateSize',
                      'BestCandidateScore', 'SomeError']
        csv_converters = {'Model':model, 'BestCandidate':model,
                          'BestCandidateScore':float_converter}
        # if log_format == 'stdout':
        #     csv_skip = 1;
        # else:
        #     csv_skip = 17;
    csv_sep = ';'
    return pd.read_csv(pathname, header=None, names=csv_header,
                       converters=csv_converters, skiprows=csv_skip,
                       skipfooter=csv_skipfooter, sep=csv_sep)


def parse_df_all(system_ids, experiment_ids, base_dir, 
                 logname_template='interactions_{0}.log', log_format='latest'):
    logname_template
    data_log = {}
    for sysid, expid in itertools.product(system_ids, experiment_ids):
        logname = logname_template.format(expid)
        pathname = os.path.sep.join([base_dir, sysid, logname])
        data_log[sysid+'_'+expid] = parse_df(pathname, log_format=log_format)
    return data_log


if __name__ == '__main__':
    pass
